package main

import (
	"context"
	"net/http"
	"os"
	"witness/api"
	"witness/conf"
	"witness/core"
	"witness/p2p"
	"witness/version"

	"github.com/prometheus/client_golang/prometheus/promhttp"
	log "github.com/sirupsen/logrus"
	"github.com/urfave/cli/v2"
)

var (
	appFlags = []cli.Flag{
		configFileFlag,
		dataDirFlag,
		metricsListenAddrFlag,
		logLevelFlag,
		grpcListenAddrFlag,
		rpcListenAddrFlag,
		privateKeyFlag,
		chainRPCFlag,
		storeContractFlag,
		witnessContractFlag,
		commitTimeFlag,
		commitExpireFlag,
	}

	p2pFlags = []cli.Flag{
		p2pUseDiscoveryFlag,
		p2pStaticPeersFlag,
		p2pBootstrapNodeAddrFlag,
		p2pLocalIPFlag,
		p2pPrivKeyFlag,
		p2pHostFlag,
		p2pTCPPortFlag,
		p2pUDPPortFlag,
		p2pMaxPeersFlag,
		p2pMaxInboundPeersFlag,
		p2pMaxOutboundPeersFlag,
		ignoreLocalIPFlag,
	}
)

func main() {
	app := cli.App{}
	app.Flags = wrapFlags(append(appFlags, p2pFlags...))
	app.Name = "witness"
	app.Usage = "this is witness"
	app.Version = version.Version
	app.Before = func(c *cli.Context) error {
		return loadFlagsFromConfig(c, app.Flags)
	}
	app.Action = func(context *cli.Context) error {
		run(context)
		return nil
	}
	if err := app.Run(os.Args); err != nil {
		log.Error(err)
	}
}

func run(ctx *cli.Context) {
	cfg := &conf.Config{
		MetricsListenAddr: ctx.String(metricsListenAddrFlag.Name),
		LogLevel:          ctx.String(logLevelFlag.Name),
		GRPCListenAddr:    ctx.String(grpcListenAddrFlag.Name),
		RPCListenAddr:     ctx.String(rpcListenAddrFlag.Name),
		PrivateKey:        ctx.String(privateKeyFlag.Name),
		ChainRPC:          ctx.String(chainRPCFlag.Name),
		StoreContract:     ctx.String(storeContractFlag.Name),
		WitnessContract:   ctx.String(witnessContractFlag.Name),
		DataDir:           ctx.String(dataDirFlag.Name),
		CommitTime:        ctx.Int(commitTimeFlag.Name),
		CommitExpire:      ctx.Int(commitExpireFlag.Name),
	}
	p2pSrv := runP2PServer(ctx)
	setLogLevel(cfg.LogLevel)
	runMetrics(cfg.MetricsListenAddr)
	w := core.RunWitness(p2pSrv, cfg)
	runGrpcServer(cfg.GRPCListenAddr, w)
	runJSONRPCServer(cfg.RPCListenAddr, w)
	select {}
}

func setLogLevel(level string) {
	_level, err := log.ParseLevel(level)
	if err != nil {
		log.WithError(err).Error("failed to parse log level")
		return
	}
	log.WithField("log level", _level).Info("set log level")
	log.SetLevel(_level)
}

func runMetrics(listen string) {
	http.Handle("/metrics", promhttp.Handler())
	log.WithField("listen", listen).Info("start prometheus metrics server")
	go func() {
		if err := http.ListenAndServe(listen, nil); err != nil {
			log.WithError(err).Error("failed to start prometheus metrics server")
		}
	}()
}

func runGrpcServer(listen string, w *core.Witness) {
	go api.StartGRPC(listen, w)
}

func runJSONRPCServer(listen string, w *core.Witness) {
	go api.StartJSONRPC(listen, w)
}

func runP2PServer(ctx *cli.Context) *p2p.Service {
	p2pCfg := &p2p.Config{
		UseDiscovery:      ctx.Bool(p2pUseDiscoveryFlag.Name),
		StaticPeers:       ctx.StringSlice(p2pStaticPeersFlag.Name),
		BootstrapNodeAddr: ctx.StringSlice(p2pBootstrapNodeAddrFlag.Name),
		LocalIP:           ctx.String(p2pLocalIPFlag.Name),
		HostAddress:       ctx.String(p2pHostFlag.Name),
		PrivateKeyPath:    ctx.String(p2pPrivKeyFlag.Name),
		TCPPort:           ctx.Uint(p2pTCPPortFlag.Name),
		UDPPort:           ctx.Uint(p2pUDPPortFlag.Name),
		MaxPeers:          ctx.Uint(p2pMaxPeersFlag.Name),
		MaxInboundPeers:   ctx.Uint(p2pMaxInboundPeersFlag.Name),
		MaxOutboundPeers:  ctx.Uint(p2pMaxOutboundPeersFlag.Name),
		IgnoreLocalIP:     ctx.Bool(ignoreLocalIPFlag.Name),
	}

	svs, err := p2p.NewService(context.Background(), p2pCfg)
	if err != nil {
		log.WithError(err).Fatal("failed to create p2p service")
	}
	svs.Start()
	return svs
}
